from typing import Callable, Union, Optional
from functools import partial

from stream2py import SourceReader, BufferReader
from stream2py.stream_buffer import StreamBuffer
from audiostream2py import PyAudioSourceReader

from taped.util import bytes_to_waveform_old

DFLT_FRM_PER_BUFFER = 2048


class VisualizationStream(SourceReader):
    def __init__(self, mk_int16_array_gen: Callable, chk_to_viz: Callable):
        self.mk_int16_array_gen = mk_int16_array_gen
        self.chk_to_viz = chk_to_viz
        self._viz_gen = None

    def _mk_viz_gen(self):
        for i, a in enumerate(self.mk_int16_array_gen()):
            yield i, self.chk_to_viz(a)

    def open(self):
        """Setup data generator"""
        self._viz_gen = self._mk_viz_gen()

    def read(self):
        return next(self._viz_gen)

    def close(self):
        """Clean up if needed"""
        del self._viz_gen
        self._viz_gen = None

    @property
    def info(self):
        """Whatever info is useful to you
        StreamBuffer will record info right after open"""
        return dict(
            mk_int16_array_gen=str(self.mk_int16_array_gen),
            chk_to_viz=str(self.chk_to_viz),
        )

    def key(self, data):
        """Convert data to a sortable value that increases with each read.
        the enumerate index in this case
        """
        return data[0]


def device_info_by_index(index):
    try:
        return next(
            d for d in PyAudioSourceReader.list_device_info() if d['index'] == index
        )
    except StopIteration:
        raise ValueError(f'Not found for input device index: {index}')


def mk_pyaudio_to_int16_array_gen_callable(audio_reader: BufferReader):
    _info = audio_reader.source_reader_info
    specific_bytes_to_wf = partial(
        bytes_to_waveform_old,
        sr=_info['rate'],
        n_channels=_info['channels'],
        sample_width=_info['width'],
    )

    def mk_pyaudio_to_int16_array_gen():
        for timestamp, wf_bytes, frame_count, time_info, status_flags in audio_reader:
            yield specific_bytes_to_wf(wf_bytes)

    return mk_pyaudio_to_int16_array_gen


class AudioStreamBuffer(StreamBuffer):
    def __init__(
        self,
        *,
        buffer_size_seconds: Union[int, float] = 60.0,
        input_device_index=None,
        sr=44100,
        width=2,
        frames_per_buffer=DFLT_FRM_PER_BUFFER,
        sleep_time_on_read_none_s: Optional[Union[int, float]] = 0.05,
        auto_drop=True,
    ):
        _info = device_info_by_index(input_device_index)
        sr = sr or int(_info['defaultSampleRate'])
        channels = _info['maxInputChannels']

        super().__init__(
            source_reader=PyAudioSourceReader(
                input_device_index=input_device_index,
                rate=sr,
                width=width,
                channels=channels,
                frames_per_buffer=frames_per_buffer,
                unsigned=False,
            ),
            maxlen=PyAudioSourceReader.audio_buffer_size_seconds_to_maxlen(
                buffer_size_seconds=buffer_size_seconds,
                rate=sr,
                frames_per_buffer=frames_per_buffer,
            ),
            sleep_time_on_read_none_s=sleep_time_on_read_none_s,
            auto_drop=auto_drop,
        )


# TODO: Reflect: Class versus function?
# def pyaudio_source_reader(
#         buffer_size_seconds: Union[int, float] = 60.0,
#         input_device_index: Optional[int] = None,
#         sr=44100,
#         width=2,
#         frames_per_buffer=2048,
#         sleep_time_on_read_none_s: Optional[Union[int, float]] = 0.05,
#         auto_drop=True
# ):
#     _info = device_info_by_index(input_device_index)
#     sr = sr or int(_info['defaultSampleRate'])
#     channels = _info['maxInputChannels']
#
#     return StreamBuffer(
#         source_reader=PyAudioSourceReader(input_device_index=input_device_index,
#                                           rate=sr,
#                                           width=width,
#                                           channels=channels,
#                                           frames_per_buffer=frames_per_buffer,
#                                           ),
#         maxlen=PyAudioSourceReader.audio_buffer_size_seconds_to_maxlen(
#             buffer_size_seconds=buffer_size_seconds,
#             rate=sr,
#             frames_per_buffer=frames_per_buffer,
#         ),
#         sleep_time_on_read_none_s=sleep_time_on_read_none_s,
#         auto_drop=auto_drop
#     )


def launch_audio_tracking(
    chk_callback: Callable,
    input_device_index: int,
    buffer_size_seconds=60,
    sr=None,
    width=2,
    frames_per_buffer=DFLT_FRM_PER_BUFFER,
):
    try:
        with AudioStreamBuffer(
            buffer_size_seconds=buffer_size_seconds,
            input_device_index=input_device_index,
            sr=sr,
            width=width,
            frames_per_buffer=frames_per_buffer,
            auto_drop=False,
            sleep_time_on_read_none_s=0.1,
        ) as audio_buffer:
            audio_reader = audio_buffer.mk_reader()

            with StreamBuffer(
                source_reader=VisualizationStream(
                    mk_int16_array_gen=mk_pyaudio_to_int16_array_gen_callable(
                        audio_reader
                    ),
                    chk_to_viz=chk_callback,
                ),
                maxlen=10,
            ) as viz_buffer:
                viz_reader = viz_buffer.mk_reader()

                yield from viz_reader
    except KeyboardInterrupt:
        print("KeyboardInterrupt: So I'm stopping the process")



"""Base of taped objects"""

from itertools import chain
from typing import Union, Optional
from collections import namedtuple
from dataclasses import dataclass

from stream2py.stream_buffer import StreamBuffer
from audiostream2py import PyAudioSourceReader
from taped.util import (
    DFLT_SR,
    DFLT_SAMPLE_WIDTH,
    DFLT_CHK_SIZE,
    DFLT_STREAM_BUF_SIZE_S,
    bytes_to_waveform,
    ensure_source_input_device_index,
)

from itertools import islice
from creek import Creek

BufferItemOutput = namedtuple(
    typename='BufferItemOutput',
    field_names=['timestamp', 'bytes', 'frame_count', 'time_info', 'status_flags'],
)


@Creek.wrap
@dataclass
class BaseBufferItems(StreamBuffer):
    """A generator of live chunks of audio bytes taken from a stream sourced from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_bytes: Sample width in bytes (1, 2, 3, or 4)
    :param sample_width: Specifies the number of frames per buffer.
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)
    """

    input_device_index: Optional[Union[int, str]] = None
    sr: int = DFLT_SR
    sample_width: int = DFLT_SAMPLE_WIDTH
    chk_size: int = DFLT_CHK_SIZE
    stream_buffer_size_s: Union[float, int] = DFLT_STREAM_BUF_SIZE_S

    def __post_init__(self):
        self.input_device_index = ensure_source_input_device_index(
            self.input_device_index
        )
        seconds_per_read = self.chk_size / self.sr

        self.maxlen = int(self.stream_buffer_size_s / seconds_per_read)
        self.source_reader = PyAudioSourceReader(
            rate=self.sr,
            width=self.sample_width,
            unsigned=True,
            input_device_index=self.input_device_index,
            frames_per_buffer=self.chk_size,
        )

        super().__init__(source_reader=self.source_reader, maxlen=self.maxlen)


class BufferItems(BaseBufferItems):
    def data_to_obj(self, data):
        return BufferItemOutput(*data)


class ByteChunks(BufferItems):
    def data_to_obj(self, data):
        return super().data_to_obj(data).bytes


# TODO: use more stable and flexible bytes_to_waveform
class WfChunks(ByteChunks):
    sample_width = 2

    def data_to_obj(self, data):
        data = super().data_to_obj(data)
        return bytes_to_waveform(data, sample_width=self.sample_width)


# TODO: use more stable and flexible bytes_to_waveform
class LiveWf(WfChunks):
    def post_iter(self, obj):
        return chain.from_iterable(obj)

    @property
    def available_samples_in_buffer(self):
        """Returns the actual max number of samples available.
        This is maxlen * chk_size when buffer is full.
        """

        return len(self.source_buffer._buffer) * self.chk_size

    def __getitem__(self, item):
        if not isinstance(item, slice):
            item = slice(item, item + 1)  # to emulate usual list[i] interface
        item = positive_slice_version(item, self.available_samples_in_buffer)
        return list(islice(self, item.start, item.stop, item.step))


def positive_slice_version(slice_, sliced_obj_len):
    """Returns a slice where negative start and stops are replaced with their positive correspondence.
    This is because `itertools.islice` doesn't accept negative entries, so we need to tell it
    explicitly what we mean by -3 (we mean sliced_obj_len - 3).

    >>> positive_slice_version(slice(-7, -1, 3), 10)
    slice(3, 9, 3)
    >>> positive_slice_version(slice(2, -2), 7)
    slice(2, 5, None)
    """

    def positivize(x):
        if x is not None and x < 0:
            x += sliced_obj_len
        return x
        # return x if x is None or x > 0 else 0

    return slice(positivize(slice_.start), positivize(slice_.stop), slice_.step)



"""Simple audio acquisition"""

from taped.display import disp_wf
from taped.base import (
    BaseBufferItems,
    BufferItems,
    ByteChunks,
    WfChunks,
    LiveWf,
    BufferItemOutput,
)
from taped.util import (
    list_recording_device_index_names,
    find_a_default_input_device_index,
    simple_chunker,
    rechunker,
    chunk_indices,
)



"""Utils for taped"""
from functools import partial
import functools
from io import BytesIO
from itertools import chain
from typing import Iterable, Union, Callable, List, Tuple

import numpy as np
import soundfile as sf
from audiostream2py import PyAudioSourceReader, get_input_device_index

# TODO: (wish) service this will builtins only


# DFLT_INPUT_DEVICE_INDEX = find_a_default_input_device_index()
DFLT_SR = 44100
DFLT_N_CHANNELS = 1
DFLT_SAMPLE_WIDTH = 2
DFLT_CHK_SIZE = 1024 * 4
DFLT_STREAM_BUF_SIZE_S = 60
read_kwargs_for_sample_width = {
    2: dict(format='RAW', subtype='PCM_16', dtype='int16'),
    3: dict(format='RAW', subtype='PCM_24'),  # what dtype?
    4: dict(format='RAW', subtype='PCM_32', dtype='int32'),
}

# monkey patching WRAPPER_ASSIGNMENTS to get "proper" wrapping (adding defaults and kwdefaults

wrapper_assignments = (
    '__module__',
    '__name__',
    '__qualname__',
    '__doc__',
    '__annotations__',
    '__defaults__',
    '__kwdefaults__',
)

update_wrapper = functools.update_wrapper
update_wrapper.__defaults__ = (functools.WRAPPER_ASSIGNMENTS, functools.WRAPPER_UPDATES)
wraps = functools.wraps
wraps.__defaults__ = (functools.WRAPPER_ASSIGNMENTS, functools.WRAPPER_UPDATES)


def bytes_to_waveform(
    b, sr=DFLT_SR, n_channels=DFLT_N_CHANNELS, sample_width=DFLT_SAMPLE_WIDTH
):
    """
    Convert raw bytes to a numpy array cast to dtype

    :param b: bytes
    :param sr: sample rate
    :param n_channels: number of channels
    :param sample_width: sample byte width [2, 3, 4]

    :return: numpy.array or ints
    """
    return sf.read(
        BytesIO(b),
        samplerate=sr,
        channels=n_channels,
        **read_kwargs_for_sample_width[sample_width],
    )[0]


def waveform_to_bytes(wf, sr=DFLT_SR, sample_width=DFLT_SAMPLE_WIDTH):
    """
    Convert raw bytes to a numpy array cast to dtype

    :param wf: iterable of ints
    :param sr: sample rate
    :param n_channels: number of channels
    :param sample_width: sample byte width [2, 3, 4]

    :return: bytes
    """
    b = BytesIO()
    subtype = read_kwargs_for_sample_width[sample_width]['subtype']
    sf.write(b, wf, samplerate=sr, format='RAW', subtype=subtype)
    b.seek(0)
    return b.read()


def bytes_to_waveform_old(
    b: bytes, sr: int, n_channels: int, sample_width: int, dtype='int16'
) -> np.array:
    """Convert raw bytes to a numpy array cast to dtype

    :param b: bytes
    :param sr: sample rate
    :param n_channels: number of channels
    :param sample_width: sample byte width [2, 3, 4]
    :param dtype: data type used by numpy, i.e. dtype=np.int16 is the same as dtype='int16'
    :return: numpy.array
    """
    sample_width_to_subtype = {
        2: 'PCM_16',
        3: 'PCM_24',
        4: 'PCM_32',
    }
    return sf.read(
        BytesIO(b),
        samplerate=sr,
        channels=n_channels,
        format='RAW',
        subtype=sample_width_to_subtype[sample_width],
        dtype=dtype,
    )[0]


def list_recording_device_index_names() -> List[Tuple[int, str]]:
    """List (index, name) of available recording devices"""
    return sorted(
        (d['index'], d['name'])
        for d in PyAudioSourceReader.list_device_info()
        if d['maxInputChannels'] > 0
    )


def find_a_default_input_device_index(verbose=True):
    return get_input_device_index(verbose=verbose)


def find_a_device_index(filt='microphone', dflt=None):
    if isinstance(filt, str):
        match_str = filt

        def filt(x):
            return match_str in x.get('name', match_str).lower()

    match = next(filter(filt, PyAudioSourceReader.list_device_info()), None)
    return (match is not None and match['index']) or dflt


def ensure_source_input_device_index(input_device_index=None):
    if input_device_index is None:
        input_device_index = find_a_default_input_device_index()
    if input_device_index is not None:
        if isinstance(input_device_index, int):
            return input_device_index
        elif isinstance(input_device_index, str):
            input_name = input_device_index
            for index, name in list_recording_device_index_names():
                if name == input_name:
                    return index
            raise ValueError(
                f'name not found in list of recording devices: {input_name}'
            )
        elif isinstance(input_device_index, tuple) and len(input_device_index) == 2:
            index, name = input_device_index
            assert isinstance(
                index, int
            ), f'expecting first element of tuple to be an int: {input_device_index}'
            assert isinstance(
                name, str
            ), f'expecting second element of tuple to be a string: {input_device_index}'
            return index
        else:
            raise ValueError(
                f"couldn't resolve input_device_index: {input_device_index}"
            )
    else:
        # TODO: Nicer way to print info (perhaps only relevant info, formated as table)
        print(
            'Need a valid input_device_index. Calling live_audio_chks.list_device_info() to information about the '
            'devices I can detect:\n'
        )
        for item in PyAudioSourceReader.list_device_info():
            print(item)
            print('')
        print(
            "---> Look in the list above and choose an input_device_index (it's called index in the printout above) "
            'that seems to be right for you!'
        )
        raise ValueError('Need a valid input_device_index')


def simple_chunker(a: Iterable, chk_size: int):
    """Generate fixed sized non-overlapping chunks of an iterable ``a``.

    >>> list(simple_chunker(range(7), 3))
    [(0, 1, 2), (3, 4, 5)]

    Most of the time, you'll want to fix the parameters of the chunker like this:

    >>> from functools import partial
    >>> chunker = partial(simple_chunker, chk_size=3)
    >>> list(chunker(range(7)))
    [(0, 1, 2), (3, 4, 5)]

    Note, the type of the chunks is always tuples, but you can easily change that using ``map``.
    For example, to change the type to be list:

    >>> list(map(list, chunker(range(7))))
    [[0, 1, 2], [3, 4, 5]]

    >>> a = range(6)
    >>> list(simple_chunker(a, 3))
    [(0, 1, 2), (3, 4, 5)]
    >>> list(simple_chunker(a, 2))
    [(0, 1), (2, 3), (4, 5)]
    >>> list(simple_chunker(a, 1))
    [(0,), (1,), (2,), (3,), (4,), (5,)]

    """
    return zip(*([iter(a)] * chk_size))


def rechunker(chks: Iterable[Iterable], chunker: Union[Callable, int]):
    """Generate fixed sized non-overlapping chunks of an iterable of chunks.
    That is, the rechunker applies a chunker to an unraveled stream of chunks,
    or more generally of iterables since they can be of varied sizes and types.

    >>> from functools import partial
    >>> chunker = partial(simple_chunker, chk_size=3)
    >>> chks = [[0], (1, 2, 3), [4, 5], iter((6, 7))]  # iterable of (different types of) iterables
    >>> list(rechunker(chks, chunker))
    [(0, 1, 2), (3, 4, 5)]

    """
    if isinstance(
        chunker, int
    ):  # if chunker is an int, take it to be a the chk_size of a simple_chunker
        chk_size = chunker
        chunker = partial(simple_chunker, chk_size=chk_size)
    yield from chunker(chain.from_iterable(chks))


from itertools import count, islice

# TODO: Restricted by integer: Need float and decimal versions
def chunk_indices(chk_size, chk_step=None, start_idx=0, end_idx=None):
    """Yields upper and integer lower bounds of integer size and step chunks.

    >>> it = chunk_indices(5)
    >>> next(it)
    (0, 5)
    >>> next(it)
    (5, 10)
    >>> list(chunk_indices(chk_size=5, chk_step=3, start_idx=0, end_idx=15))
    [(0, 5), (3, 8), (6, 11), (9, 14)]
    """
    chk_step = chk_step or chk_size
    bt_start = start_idx
    tt_start = bt_start + chk_size
    chk_index_iterator = zip(count(start_idx, chk_step), count(tt_start, chk_step))
    if end_idx is None:
        # iterate forever
        return chk_index_iterator
    else:  # end_idx (the maximum that tt_start should have) is not None
        # so we want a finite number of iterators, so we computer the num of iterations
        n_iterations = 1 + int((end_idx - bt_start - chk_size) / chk_step)  # verify!
        return islice(chk_index_iterator, n_iterations)



"""
Recording audio to wav file
---------------------------
This example requires PyAudio
::
    pip install PyAudio

.. autoclass:: stream2py.examples.usage.record_audio_to_files.BufferReaderConsumer()
    :members:
    :show-inheritance:
    :undoc-members:

    .. automethod:: __init__

.. autoclass:: stream2py.examples.usage.record_audio_to_files.PyAudioSaver()
    :members:
    :show-inheritance:
    :undoc-members:

    .. automethod:: __init__

.. autofunction:: stream2py.examples.usage.record_audio_to_files.audio_to_files

"""

from abc import ABCMeta, abstractmethod

from contextlib import suppress
import logging
import os
import threading
import time
import wave
from typing import Union

from stream2py import StreamBuffer, BufferReader
from audiostream2py import PyAudioSourceReader, PaStatusFlags


logger = logging.getLogger(__name__)


class BufferReaderConsumer(threading.Thread, metaclass=ABCMeta):
    """Call reader_handler function with reader at defined time intervals between calls"""

    def __init__(
        self,
        buffer_reader: BufferReader,
        interval: Union[int, float],
        logging_enabled: bool = False,
    ):
        """

        :param buffer_reader: BufferReader created from a StreamBuffer
        :param interval: seconds between reader_handler calls
        :param logging_enabled: log debug messages using logging module
        """
        threading.Thread.__init__(self, daemon=True)
        assert isinstance(buffer_reader, BufferReader)
        self.buffer_reader = buffer_reader
        self.interval = interval
        self.stop_event = threading.Event()
        self.logging_enabled = logging_enabled

    def stop(self):
        """Sets stop event and waits for thread to finish"""
        self.stop_event.set()
        if self.logging_enabled:
            logger.debug(f'Consumer stopping... {self.__class__.__name__}')
        self.join()

    def run(self):
        """Calls self.reader_handler in a loop while sleeping for self.interval seconds after each call until stop event
        is set or Exception is raised by self.reader_handler"""
        if self.logging_enabled:
            logger.debug(f'Consumer starting! {self.__class__.__name__}')
        try:
            while not self.stop_event.is_set():
                self.reader_handler(self.buffer_reader)
                time.sleep(self.interval)
        finally:
            with suppress(Exception):
                self.stop()
        if self.logging_enabled:
            logger.debug(f'Consumer stopped! {self.__class__.__name__}')

    @abstractmethod
    def reader_handler(self, buffer_reader: BufferReader):
        """Implements how to read from buffer and what to do with the data"""
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class PyAudioSaver(BufferReaderConsumer):
    """Save audio to wav file. Start a new wav file on errors such as input overflow."""

    def __init__(self, buffer_reader, interval, rootdir, logging_enabled):
        """

        :param buffer_reader: BufferReader created from a StreamBuffer with PyAudioSourceReader
        :param interval: seconds between reader_handler calls
        :param rootdir: folder path where files will be saved under
        :param logging_enabled: log debug messages using logging module
        """
        super().__init__(buffer_reader, interval, logging_enabled)

        self.session_data_path_format = os.path.join(
            os.path.expanduser(rootdir), '{session}', 'd'
        )
        self.path_format = os.path.join(
            self.session_data_path_format, '{timestamp}.wav'
        )
        self.error_path_format = os.path.join(
            self.session_data_path_format, '{timestamp}_ERROR_{status_flags}.wav'
        )
        self.file = None

    def reader_handler(self, buffer_reader: BufferReader):
        new_data_list = buffer_reader.range(
            start=0, stop=float('inf'), ignore_no_item_found=True, only_new_items=True
        )
        if new_data_list is not None:
            for (
                timestamp,
                in_data,
                frame_count,
                time_info,
                status_flags,
            ) in new_data_list:
                if PaStatusFlags(status_flags) != PaStatusFlags.paNoError:
                    if self.logging_enabled:
                        logger.debug(PaStatusFlags(status_flags))
                    # an error occurred, close file
                    self.close_file()
                    # save to error file
                    self.save_error(
                        buffer_reader.source_reader_info,
                        timestamp,
                        in_data,
                        status_flags,
                    )
                else:
                    if self.file_is_open() is False:
                        self.open_file(buffer_reader.source_reader_info, timestamp)
                    self.write_to_file(in_data)

    def file_is_open(self) -> bool:
        """
        Check if file is open
        :return: boolean
        """
        return self.file is not None

    def open_file(self, source_reader_info: dict, timestamp: int) -> wave.Wave_write:
        """open wav file for writing"""
        file_path = self.path_format.format(
            session=source_reader_info['bt'], timestamp=timestamp
        )
        if self.logging_enabled:
            logger.debug(f'opening file: {file_path}')
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        self.file = wave.open(file_path, mode='wb')
        self.set_params(self.file, source_reader_info)
        return self.file

    @staticmethod
    def set_params(file: wave.Wave_write, source_reader_info: dict):
        """Set wav file params"""
        nchannels = source_reader_info['channels']
        sampwidth = source_reader_info['width']
        framerate = source_reader_info['rate']
        nframes = 0
        comptype = 'NONE'
        compname = 'not compressed'
        file.setparams((nchannels, sampwidth, framerate, nframes, comptype, compname))

    def write_to_file(self, in_data):
        """Write to wav file"""
        self.file.writeframes(in_data)

    def close_file(self):
        """Close wav file"""
        self.file.close()
        self.file = None

    def stop(self):
        """Stop BufferReaderConsumer loop and close wav file"""
        super().stop()
        if self.file_is_open() is True:
            self.close_file()

    def save_error(self, source_reader_info, timestamp, in_data, status_flags):
        """Save to error wav file"""
        file_path = self.error_path_format.format(
            session=source_reader_info['bt'],
            timestamp=timestamp,
            status_flags=status_flags,
        )
        if self.logging_enabled:
            logger.debug(f'saving error file: {file_path}')
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        error_file = wave.open(file_path, mode='wb')
        self.set_params(error_file, source_reader_info)
        error_file.writeframes(in_data)
        error_file.close()


def audio_to_files(
    rate,
    width,
    channels,
    input_device_index,
    frames_per_buffer,
    interval,
    rootdir,
    logging_enabled,
):
    """Basically the main function to run the example.
    It will record audio with audiostream2py.PyAudioSourceReader
    and save to wav files with stream2py.examples.usage.record_audio_to_files.PyAudioSaver

    Check this source code to see how to put together the three components: SourceReader, StreamBuffer, BufferReader

    :param rate: Specifies the desired sample rate (in Hz)
    :param width: Sample width in bytes (1, 2, 3, or 4)
    :param channels: The desired number of input channels. Ignored if input_device is not specified (or None).
    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param frames_per_buffer: Specifies the number of frames per buffer.
    :param interval: seconds between reader_handler calls
    :param rootdir: folder path where files will be saved under
    :param logging_enabled: log debug messages using logging module
    """
    from stream2py.utility.logger import set_logging_config

    set_logging_config(level=logging.DEBUG)

    seconds_per_read = frames_per_buffer / rate
    seconds_to_keep_in_stream_buffer = 60

    maxlen = int(seconds_to_keep_in_stream_buffer / seconds_per_read)
    source_reader = PyAudioSourceReader(
        rate=rate,
        width=width,
        channels=channels,
        unsigned=True,
        input_device_index=input_device_index,
        frames_per_buffer=frames_per_buffer,
    )

    with StreamBuffer(source_reader=source_reader, maxlen=maxlen) as stream_buffer:
        """keep open and save to file until stop event"""
        buffer_reader = stream_buffer.mk_reader()
        with PyAudioSaver(
            buffer_reader,
            interval=interval,
            rootdir=rootdir,
            logging_enabled=logging_enabled,
        ) as pasave:
            try:
                pasave.join()
            except KeyboardInterrupt:
                pass


audio_to_files.list_device_info = lambda: PyAudioSourceReader.list_device_info()

if __name__ == '__main__':
    audio_to_files(
        rate=44100,
        width=2,
        channels=1,
        input_device_index=6,
        frames_per_buffer=1024 * 4,
        interval=1,
        rootdir='~/odir/stream2py',
        logging_enabled=True,
    )



from functools import partial
from itertools import chain
from contextlib import contextmanager
from typing import Iterable, Union, Callable

from stream2py.stream_buffer import StreamBuffer
from audiostream2py import PyAudioSourceReader
from taped.util import (
    DFLT_SR,
    DFLT_SAMPLE_WIDTH,
    DFLT_CHK_SIZE,
    DFLT_STREAM_BUF_SIZE_S,
    bytes_to_waveform,
    ensure_source_input_device_index,
)


@contextmanager
def live_audio_chks(
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
):
    """A generator of live chunks of audio bytes taken from a stream sourced from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_bytes: Sample width in bytes (1, 2, 3, or 4)
    :param sample_width: Specifies the number of frames per buffer.
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)
    """
    input_device_index = ensure_source_input_device_index(input_device_index)

    seconds_per_read = chk_size / sr

    maxlen = int(stream_buffer_size_s / seconds_per_read)
    # print(maxlen)
    source_reader = PyAudioSourceReader(
        rate=sr,
        width=sample_width,
        unsigned=True,
        input_device_index=input_device_index,
        frames_per_buffer=chk_size,
    )

    # _bytes_to_waveform = partial(audio_pokes_version_of_bytes_to_waveform, sr=sr,
    #                              sample_width=sample_width)
    with StreamBuffer(source_reader=source_reader, maxlen=maxlen) as stream_buffer:
        """keep open and save to file until stop event"""
        yield iter(stream_buffer)


live_audio_chks.list_device_info = PyAudioSourceReader.list_device_info


# TODO: live_wf_ctx and live_wf: Lot's of repeated code. Address this.
@contextmanager
def live_wf_ctx(
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
):
    """A context manager providing a generator of live waveform sample values taken from a stream sourced
    from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_width: Sample width in bytes (1, 2, 3, or 4)
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)

    >>> from time import sleep
    >>> from itertools import islice
    >>> # enter the id of your microphone and get a live waveform source!
    >>> # (if None, will try to figure it out)
    >>> with live_wf_ctx(input_device_index=None) as wf_gen:
    ...
    ...     # Now wait a bit, say some silly things, then ask for a few samples...
    ...     sleep(1.1)
    ...     wf = list(islice(wf_gen, 0, 44100 * 1))
    >>> # and now listen to that wf and be embarrassed...
    >>> # ... or just look at the size (less fun though)
    >>> len(wf)
    44100
    """
    with live_audio_chks(
        input_device_index=input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    ) as live_audio_chunks:
        _bytes_to_waveform = partial(
            bytes_to_waveform, sr=sr, sample_width=sample_width
        )
        yield chain.from_iterable(
            map(lambda x: _bytes_to_waveform(x[1]), live_audio_chunks)
        )
    live_audio_chunks.close()


from itertools import islice
from dataclasses import dataclass


@dataclass
class LiveAudioChks(StreamBuffer):
    """A generator of live chunks of audio bytes taken from a stream sourced from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_bytes: Sample width in bytes (1, 2, 3, or 4)
    :param sample_width: Specifies the number of frames per buffer.
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)
    """

    input_device_index = None
    sr = DFLT_SR
    sample_width = DFLT_SAMPLE_WIDTH
    chk_size = DFLT_CHK_SIZE
    stream_buffer_size_s = DFLT_STREAM_BUF_SIZE_S

    def __post_init__(self):
        self.input_device_index = ensure_source_input_device_index(
            self.input_device_index
        )
        seconds_per_read = self.chk_size / self.sr

        self.maxlen = int(self.stream_buffer_size_s / seconds_per_read)
        self.source_reader = PyAudioSourceReader(
            rate=self.sr,
            width=self.sample_width,
            unsigned=True,
            input_device_index=self.input_device_index,
            frames_per_buffer=self.chk_size,
        )

        super().__init__(source_reader=self.source_reader, maxlen=self.maxlen)

    def __next__(self):
        return next(iter(self))


class LiveWf(LiveAudioChks):
    def __enter__(self):
        super().__enter__()
        self._bytes_to_waveform = partial(
            bytes_to_waveform, sr=self.sr, sample_width=self.sample_width
        )
        self.live_wf = chain.from_iterable(
            map(lambda x: self._bytes_to_waveform(x[1]), self)
        )
        return self

    # TODO: Protect from using this, or make it work
    # def __iter__(self):
    #     raise NotImplementedError("")
    #     # yield from self.live_wf

    def __getitem__(self, item):
        if not isinstance(item, slice):
            item = slice(item, item + 1)  # to emulate usual list[i] interface
        return list(islice(self.live_wf, item.start, item.stop, item.step))

    def __exit__(self, *args, **kwargs):
        return super().__exit__(*args, **kwargs)


def live_wf(
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
):
    """A generator of live waveform sample values taken from a stream sourced from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_width: Sample width in bytes (1, 2, 3, or 4)
    :param n_channels: The desired number of input channels. Ignored if input_device is not specified (or None).
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)

    >>> from time import sleep
    >>> from itertools import islice
    >>> # enter the id of your microphone and get a live waveform source!
    >>> # (if None, will try to figure it out)
    >>> wf_gen = live_wf(input_device_index=None)
    >>>
    >>> # Now wait a bit, say some silly things, then ask for a few samples...
    >>> sleep(1.2)
    >>> wf = list(islice(wf_gen, 0, 44100 * 1))
    >>> # and now listen to that wf and be embarrassed...
    >>> # ... or just look at the size (less fun though)
    >>> len(wf)
    44100

    Don't forget to close! (or use live_wf_ctx context manager).
    >>> wf_gen.close()

    After wf_gen is closed, you can still ask it for data.
    It just won't give you any.
    >>> wf = list(islice(wf_gen, 0, 44100 * 1))
    >>> len(wf)
    0

    Here wf_gen is a generator, so closing means: https://docs.python.org/2.5/whatsnew/pep-342.html
    """
    # TODO: Find a way to copy from containing function's signature and calling LiveAudioChunks with that
    with live_wf_ctx(
        input_device_index=input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    ) as live_wf:
        yield from live_wf


live_wf.list_device_info = PyAudioSourceReader.list_device_info


def simple_chunker(a: Iterable, chk_size: int):
    """Generate fixed sized non-overlapping chunks of an iterable ``a``.

    >>> list(simple_chunker(range(7), 3))
    [(0, 1, 2), (3, 4, 5)]

    Most of the time, you'll want to fix the parameters of the chunker like this:

    >>> from functools import partial
    >>> chunker = partial(simple_chunker, chk_size=3)
    >>> list(chunker(range(7)))
    [(0, 1, 2), (3, 4, 5)]

    Note, the type of the chunks is always tuples, but you can easily change that using ``map``.
    For example, to change the type to be list:

    >>> list(map(list, chunker(range(7))))
    [[0, 1, 2], [3, 4, 5]]

    >>> a = range(6)
    >>> list(simple_chunker(a, 3))
    [(0, 1, 2), (3, 4, 5)]
    >>> list(simple_chunker(a, 2))
    [(0, 1), (2, 3), (4, 5)]
    >>> list(simple_chunker(a, 1))
    [(0,), (1,), (2,), (3,), (4,), (5,)]

    """
    return zip(*([iter(a)] * chk_size))


def rechunker(chks: Iterable[Iterable], chunker: Union[Callable, int]):
    """Generate fixed sized non-overlapping chunks of an iterable of chunks.
    That is, the rechunker applies a chunker to an unraveled stream of chunks,
    or more generally of iterables since they can be of varied sizes and types.

    >>> from functools import partial
    >>> chunker = partial(simple_chunker, chk_size=3)
    >>> chks = [[0], (1, 2, 3), [4, 5], iter((6, 7))]  # iterable of (different types of) iterables
    >>> list(rechunker(chks, chunker))
    [(0, 1, 2), (3, 4, 5)]

    """
    if isinstance(
        chunker, int
    ):  # if chunker is an int, take it to be a the chk_size of a simple_chunker
        chk_size = chunker
        chunker = partial(simple_chunker, chk_size)
    yield from chunker(chain.from_iterable(chks))


def record_some_sound(
    save_to_file,
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
    verbose=True,
):
    def get_write_file_stream():
        if isinstance(save_to_file, str):
            return open(save_to_file, 'wb')
        else:
            return save_to_file  # assume it's already a stream

    def clog(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    seconds_per_read = chk_size / sr
    maxlen = int(stream_buffer_size_s / seconds_per_read)

    source_reader = PyAudioSourceReader(
        rate=sr,
        width=sample_width,
        unsigned=True,
        input_device_index=input_device_index,
        frames_per_buffer=chk_size,
    )

    with StreamBuffer(source_reader=source_reader, maxlen=maxlen) as stream_buffer:
        """keep open and save to file until stop event"""
        clog('starting the recording...')
        with get_write_file_stream() as write_stream:
            while True:
                try:
                    chk = source_reader.read()
                    print(type(chk), len(chk))
                except KeyboardInterrupt:
                    clog('stopping the recording...')
                    break

    clog('Done.')



from taped.base import BaseBufferItems
from taped.util import DFLT_SR, DFLT_SAMPLE_WIDTH, DFLT_CHK_SIZE, DFLT_STREAM_BUF_SIZE_S


def record_some_sound(
    save_to_file,
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
    verbose=True,
):
    def get_write_file_stream():
        if isinstance(save_to_file, str):
            return open(save_to_file, 'wb')
        else:
            return save_to_file  # assume it's already a stream

    def clog(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    buffer_items = BaseBufferItems(
        input_device_index=input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    )
    with buffer_items:
        """keep open and save to file until stop event"""
        clog('starting the recording (you can KeyboardInterrupt at any point)...')
        with get_write_file_stream() as write_stream:
            for item in buffer_items:
                try:
                    write_stream.write(item.bytes)
                except KeyboardInterrupt:
                    clog('stopping the recording...')
                    break

    clog('Done.')



from inspect import getmodule
import matplotlib.pylab as plt
from numpy import linspace


def getmodulename(obj, default=''):
    """Get name of module of object"""
    return getattr(getmodule(obj), '__name__', default)


def plot_wf(wf, sr=None, **kwargs):
    if sr is not None:
        plt.plot(linspace(start=0, stop=len(wf) / float(sr), num=len(wf)), wf, **kwargs)
    else:
        plt.plot(wf, **kwargs)


def disp_wf(wf, sr=44100, autoplay=False, wf_plot_func=plt.specgram):
    if wf_plot_func is not None:
        if getmodulename(wf_plot_func, '').startswith('matplotlib'):
            plt.figure(figsize=(16, 5))
        wf_plot_func(wf)
    try:
        from IPython.display import Audio

        return Audio(data=wf, rate=sr, autoplay=autoplay)
    except:
        pass



from functools import partial

import numpy as np
from IPython.core.display import clear_output

from taped.scrap.live_audio import launch_audio_tracking
from taped.util import find_a_default_input_device_index

from lined.tools import iterize
from lined import Pipeline

int16_max = np.iinfo(np.int16).max


def log_max_chk_to_fv(chk: np.ndarray, gain=50):
    """A chk_to_fv that has to do with volume"""
    return [np.log(max(chk.max(), abs(chk.min()), 1)) / np.log(int16_max) * gain]
    # return np.log(min(1, max(np.abs(chk)))) / np.log(int16_max) * gain


dflt_chk_to_fv = log_max_chk_to_fv


def ap_fv_to_viz(fv):
    return f'/><{r"{" * int(sum(fv))}">'


dflt_fv_to_viz = ap_fv_to_viz


def ap_chk_to_viz(chk: np.ndarray, gain=50):
    return ap_fv_to_viz(log_max_chk_to_fv(chk, gain))


dflt_chk_to_viz = ap_chk_to_viz


class DFLT:
    gain = 10
    cutoff = 50
    dispstr = '*'
    prefix = '|'
    fv_num_sep = '\n'


def viz_norm(x, gain=10, cutoff=50):
    return min(cutoff, x * gain)


def ascii_levels(x, viz_norm=viz_norm, dispstr=DFLT.dispstr, prefix=DFLT.prefix):
    return prefix + dispstr * int(viz_norm(x))


dflt_num_to_viz_str = ascii_levels


def multipart_fv_str(
    fv, num_to_viz_str=dflt_num_to_viz_str, fv_num_sep=DFLT.fv_num_sep
):
    return fv_num_sep.join(map(num_to_viz_str, fv)) + '\n'


ascii_levels_fv_to_str = partial(
    multipart_fv_str, num_to_viz_str=ascii_levels, fv_num_sep=DFLT.fv_num_sep
)

raw_fv_to_str = partial(
    multipart_fv_str, num_to_viz_str=str, fv_num_sep=DFLT.fv_num_sep
)

dflt_fv_to_str = ascii_levels_fv_to_str


def chk_to_fv_viz(chk: np.ndarray, chk_to_fv=dflt_chk_to_fv, fv_to_str=dflt_fv_to_str):
    """Takes a chk and returns a string representing the fv. Meant to be displayed dynamically"""
    fv = chk_to_fv(chk)
    fv_str = fv_to_str(fv)
    return fv_str


def refresh_and_print(iterator_output):
    clear_output(wait=True)
    n, viz = iterator_output
    print(n, viz, sep='\n')


from collections import deque


class CallableDeque(deque):
    def __call__(self, x):
        self.append(x)


dflt_output_callback = refresh_and_print


def make_and_consume_iterator(
    iterator_func, *iterator_func_args, **iterator_func_kwargs
):
    """make an consume and iterator"""
    for _ in iterator_func(*iterator_func_args, **iterator_func_kwargs):
        pass


def mk_launcher(
    chk_to_fv=dflt_chk_to_fv,
    fv_to_viz=dflt_fv_to_viz,
    input_device_index=None,
    output_callback=dflt_output_callback,
):
    if input_device_index is None:
        input_device_index = find_a_default_input_device_index()

    def chk_to_fv_viz(chk):
        return fv_to_viz(chk_to_fv(chk))

    pipe = Pipeline(launch_audio_tracking, iterize(output_callback))

    def launch():
        make_and_consume_iterator(
            pipe, chk_callback=chk_to_fv_viz, input_device_index=input_device_index
        )

    return launch
